function TerminalRadonizer(varargin)
% Performs Radonizer calculations without the UI. Optional arguments specified in parser format.
% TerminalRadonizer('source', stringVar) opens file at stringVar path.
% TerminalRadonizer('savepath', stringVar) saves the result at stringVar path.
% TerminalRadonizer('method', stringVar) performs Brute Force if stringVar is equal to 'bruteforce'; otherwise Particle Swarm is used.
% Arguments can be combined. If no path arguments are given, UI popups to specify them are invoked.
%% Parse input
    parser = inputParser;
    parser.addParameter('source','_undefined',@isfile);
    parser.addParameter('savepath','_undefined',@isfolder);
    parser.addParameter('method','swarm',@isstr);
    parser.parse(varargin{:});

    if strcmp(parser.Results.source,'_undefined')
        [fileStr,pathStr] = uigetfile({'*.rn3;*.mat', 'Radonizer project (*.rn3, *.mat)'},'Choose Radon project');
        fullfileStr = fullfile(pathStr,fileStr);
        clear('fileStr', 'pathStr');
    else
        fullfileStr = parser.Results.source;
    end
    
    if strcmp(parser.Results.savepath,'_undefined')
        [fileStr,pathStr] = uiputfile({'*.rn3';'*.mat'},'Choose project save location...','radonProject');
        savepathStr = fullfile(pathStr,fileStr);
        clear('fileStr', 'pathStr');
    else
        savepathStr = parser.Results.savepath;
    end
   
    if strcmp(parser.Results.method,'bruteforce')
        method = 'bruteforce';
    else
        method = 'swarm';
    end
    
    % Clearing variables
    clear('varargin','parser');
    
 %% Load the files
    try
        loaded = load(fullfileStr,'-mat');
    catch
        error('Path to source file or the file itself is invalid! Quitting Terminal Radonizer...');
    end

    if ~isfile(loaded.Files.ft3)
        [fileFT3,pathFT3] = uigetfile({'*.ft3','NMRPipe 3D spectrum (*.ft3)'}, sprintf('%s not found', loaded.Files.ft3));
        if isequal(fileFT3,0) || isequal(pathFT3,0)
            error('Program terminated by user action!');
        end
        loaded.Files.ft3 = fullfile(pathFT3,fileFT3);
    end
%% Prepare data
    app = struct();
    % app.FilePaths.ft3 = loaded.Files.ft3;
    [spec, headerInfo] = import_NMRPipe_FT(loaded.Files.ft3);
    app.Data.init_spectrum = spec;
    app.Data.spectrum = spec;
    app.Data.ifftn = ifftn(spec);
    app.Data.headerInfo = headerInfo;
    % app.Data.tempList = loaded.AppData.tempList;
    app.Data.NUSList = loaded.AppData.NUSList;
%     app.Displayed.spectrum = real(fftn(app.Data.ifftn)); % sanity check in case the spectrum was given as abs
%     app.Displayed.projection = cell(3,1);
%     app.Displayed.limits = [ones(3,1) size(spec)'];
%     app.Displayed.limitsPts = app.Displayed.limits;
    % app.resetWorkspace();
    if isfield(loaded, 'LastState')
        if isfield(loaded.LastState, 'PeakTable')
            app.UITable.Data = loaded.LastState.PeakTable;
        end
%         if isfield(loaded.LastState, 'SwarmParams')
%             app.ParticlesSpinner.Value = loaded.LastState.SwarmParams(1);
%             app.IterationsSpinner.Value = loaded.LastState.SwarmParams(2);
%             app.RunsSpinner.Value = loaded.LastState.SwarmParams(3);
%         end
%         if isfield(loaded.LastState, 'LimitedSwarm')
%             app.LimitedSwarmCheckBox.Value = loaded.LastState.LimitedSwarm;
%         end
%         if isfield(loaded.LastState, 'LimitedSwarmType')
%             app.RangeTypeSwitch.Value = loaded.LastState.LimitedSwarmType;
%         end
%         if isfield(loaded.LastState, 'LimitedSwarmDeltas')
%             app.dQ1Button.Value = loaded.LastState.LimitedSwarmDeltas(1);
%             app.dQ2Button.Value = loaded.LastState.LimitedSwarmDeltas(2);
%             app.dQ3Button.Value = loaded.LastState.LimitedSwarmDeltas(3);
%         end
%         if isfield(loaded.LastState, 'LimitedSwarmParams')
%             app.dQ1SwarmRangeEditField.Value = loaded.LastState.LimitedSwarmParams(1);
%             app.dQ2SwarmRangeEditField.Value = loaded.LastState.LimitedSwarmParams(2);
%             app.dQ3SwarmRangeEditField.Value = loaded.LastState.LimitedSwarmParams(3);
%         end
        if isfield(loaded.LastState, 'Processing')
            % app.ProcessingCheckBoxValueChanged();
            app.Data.spectrum = app.Data.init_spectrum;
            if ~isequal(loaded.LastState.Processing, zeros(2,3))
                for d = 1:3
                    if loaded.LastState.Processing(1,d) == 1
                        app.Data.spectrum = fftshift(app.Data.spectrum,d);
                    end
                end
                for d = 1:3
                    if loaded.LastState.Processing(2,d) == 1
                        app.Data.spectrum = flip(app.Data.spectrum,d);
                    end
                end
            end
            app.Data.ifftn = ifftn(app.Data.spectrum);
            % app.deltaSpinnerValueChanged();
        end
%         if isfield(loaded.LastState, 'SearchedDeltas')
%             deltaNames = {'','',''}; % {'','','','','',''};
%             for delta = 1:3 % 1:6
%                 deltaNames{delta} = strcat(char(916),'(',app.Data.headerInfo.dim{delta}.FDFiLabel,')');
%                 % deltaNames{delta+3} = strcat(char(916),char(178),'(',app.Data.headerInfo.dim{delta}.FDFiLabel,')');
%             end
%             app.SearchedListBox.ItemsData = loaded.LastState.SearchedDeltas; app.SearchedListBox.Items = deltaNames(app.SearchedListBox.ItemsData);
%             omittedDeltas = 1:3; % 1:6;
%             omittedDeltas(ismember(omittedDeltas, loaded.LastState.SearchedDeltas)) = [];
%             app.OmittedListBox.ItemsData = omittedDeltas; app.OmittedListBox.Items = deltaNames(app.OmittedListBox.ItemsData);
%         end
    end
    % catch
    %    warndlg('Failed to load project. Check your read/write permissions for the target folder.','Project loading failed','modal');
    % end
%% Begin calculations
    if strcmp(method,'swarm')
        loaded.LastState.PeakTable = BruteForceButtonPushed(app,loaded);
    elseif strcmp(method,'bruteforce')
        loaded.LastState.PeakTable = BruteForceButtonPushed(app,loaded);
    end

%% Save results
    Files = loaded.Files; AppData = loaded.AppData; LastState = loaded.LastState;
    save(savepathStr,'Files','AppData','LastState');
    
%% Nested functions    
    function [spectrum, header] = import_NMRPipe_FT(filename, varargin)
        % Imports NMRPipe .ft2, .ft3 and .ft4 files. Returns [spectrum, header] where header is returned as a float vector.
        % NOTE: Matlab starts header byte indexing from 1 while NMRPipe's showhdr command from 0.

        % Arguments:
        % char/string 'filename': name of the file to be processed. If '' or no argument is given, file selection UI pops up.
        % logical/numeric 'showLog': if set to true/1 displays header info in the console.

        if ~exist('filename','var') || strcmp(filename,'')
            [fileString, pathString] = uigetfile('*.ft2;*.ft3;*.ft4',...
                'Select NMRPipe FT2/FT3/FT4 file to import','test.ft2');
            filename = fullfile(pathString,fileString);
        end
        % Parsing input arguments
        args = inputParser;
        addRequired(args,'filename',@(x) isstring(x) || ischar(x));
        addOptional(args,'parseHeader',true)
        parse(args,filename,varargin{:});
        % Loading the data.
        [rawFT, msg] = fopen(args.Results.filename);
        if rawFT == -1
            error('Unable to load %s: %s', args.Results.filename,msg);
            return
        end
        % Separating header from FT-spectrum.
        data = fread(rawFT,'single');
        fclose(rawFT);
        header = data(1:512);
        no_header = data(513:end);
        if header(10) > 1 % header(10) = number of dimensions
            FT_dim = [header(99+1), header(219+1), header(15+1), header(32+1)];
            spectrum = reshape(no_header, FT_dim(1:header(9+1)));
        end
        % Return already parsed header instead of raw values (default setting).
        if args.Results.parseHeader == true
            info = struct();
            info.FDMagic = header(0+1);
            info.FDFLTFormat = header(1+1);
            info.FDFLOrder = header(2+1);
            info.FDSize = header(99+1);
            info.FDRealSize = header(97+1);
            info.FDTransposed = header(221+1);
            info.FDSpecNum = header(219+1);
            info.FDQuadFlag = header(106+1);
            info.FDMax = header(247+1);
            info.FDMin = header(248+1);
            info.FDScaleFlag = header(250+1);
            info.FDDispMax = header(251+1);
            info.FDDispMin = header(252+1);
            info.FDPipeFlag = header(57+1);
            info.FDSliceCount = header(443+1);
            info.FDFileCount = header(442+1);
            info.FDPipeCount = header(75+1);
            info.FDCARFlags = header(76+1);
            info.FDFirstPlane = header(77+1);
            info.FDLastPlane = header(78+1);
            info.FDDimCount = header(9+1);
            info.FDPlaneLoc = header(14+1);
            info.FDAQSign = header(64+1);
            info.FDPartition = header(65+1);
            info.FDUser1 = header(70+1);
            info.FDUser2 = header(71+1);
            info.FDUser3 = header(72+1);
            info.FDUser4 = header(73+1);
            info.FDUser5 = header(74+1);
            info.FDMCFlag = header(135+1);
            info.FDLastBlock = header(359+1);
            info.FDContBlock = header(360+1);
            info.FDBaseBlock = header(361+1);
            info.FDPeakBlock = header(362+1);
            info.FDBMAPBlock = header(363+1);
            info.FDHistBlock = header(364+1);
            info.FD1DBlock = header(365+1);

            info.FDMonth = header(294+1);
            info.FDDay = header(295+1);
            info.FDYear = header(296+1);

            info.FDNoise = header(153+1);
            info.FDRank = header(180+1);
            info.FDTemperature = header(157+1);
            info.FD2DVirgin = header(399+1);
            info.FD2DPhase = header(256+1);
            info.FDTau = header(199+1);

            info.FDSRCName = header(286+1); % 16
            info.FDUserName = header(290+1); % 16
            info.FDOperName = header(464+1); % 32
            info.FDTitle = header(297+1); % 60
            info.FDComment = header(312+1); % 160

            info.dim = cell(4,1);

            FDFiObs = [218,119,10,28]+1;  %
            FDFiSW = [229,100,11,29]+1;   %
            FDFiOrig = [249,101,12,30]+1; %
            FDFiFTFlag = [222,220,13,31]+1;
            FDFiSize = [98,96,200,201]+1;
            FDFiApod = [428,95,50,53]+1;
            FDFiQuadFlag = [55,56,51,54]+1;
            FDFiUnits = [234,152,58,59]+1;
            FDFiP0 = [245,109,60,62]+1;
            FDFiP1 = [246,110,61,63]+1;
            FDFiCAR = [67,66,68,69]+1;
            FDFiCenter = [80,79,81,82]+1;

            FDDimOrder = [24,25,26,27]+1;
            FDFiLabel = [18,16,20,22]+1;
            FDSize = [219,99,15,32]+1;    %
            dimOrder = header(FDDimOrder)';

            for d = 1:info.FDDimCount
                dim = dimOrder(d);
                info.dim{d} = struct();
                info.dim{d}.FDFiObs = header(FDFiObs(dim));
                info.dim{d}.FDFiSW = header(FDFiSW(dim));
                info.dim{d}.FDFiOrig = header(FDFiOrig(dim));
                info.dim{d}.FDFiFTFlag = header(FDFiFTFlag(dim));
                info.dim{d}.FDFiSize = header(FDFiSize(dim));
                info.dim{d}.FDFiApod = header(FDFiApod(dim));
                info.dim{d}.FDFiQuadFlag = header(FDFiQuadFlag(dim));
                info.dim{d}.FDFiUnits = header(FDFiUnits(dim));
                info.dim{d}.FDFiP0 = header(FDFiP0(dim));
                info.dim{d}.FDFiP1 = header(FDFiP1(dim));
                info.dim{d}.FDFiCAR = header(FDFiCAR(dim));
                info.dim{d}.FDFiCenter = header(FDFiCenter(dim));

                info.dim{d}.FDFiLabel = generate_axis_label(header(FDFiLabel(dim))); % 8
                info.dim{d}.FDDimOrder = header(FDDimOrder(dim));
                info.dim{d}.FDSize = header(FDSize(dim));

                info.dim{d}.ppmRange = [header(FDFiOrig(dim))/header(FDFiObs(dim)), header(FDFiOrig(dim))/header(FDFiObs(dim))+header(FDFiSW(dim))/header(FDFiObs(dim))];
                info.dim{d}.ppmPerPoint = header(FDFiSW(dim))/header(FDFiObs(dim))/header(FDSize(dim));
            end
            header = info;
        end
        
        function [axis_label] = generate_axis_label(target_single)
            % Converts NMRPipe header single precision float to Matlab char.
            dec_txt = hex2dec(flip(reshape(num2hex(single(target_single)),[2,4]),2)');
            dec_txt(dec_txt == 0) = [];
            txt_temp = strcat(reshape(char(dec_txt),1,[]));
            [A_num, nucleus] = regexp(txt_temp,'^(\d+)','match','split');
            if isempty(A_num)
                axis_label = nucleus;
            else
                axis_label = strcat('^{',A_num,'}',nucleus);
            end
            axis_label = char(axis_label);
        end
    end

    function pts = ppm2pts(app, ppms, dim)
        pts = round(interp1(app.Data.headerInfo.dim{dim}.ppmRange,[1, size(app.Data.spectrum, dim)],ppms,'linear','extrap'));
    end

    function ppms = pts2ppm(app, pts, dim)
        ppms = interp1([1, size(app.Data.spectrum, dim)],app.Data.headerInfo.dim{dim}.ppmRange,pts,'linear','extrap');
    end

    function [result] = trim3D(target,limits)
        for dim = 1:3
            limits(dim,1) = min([size(target,dim), max([1, limits(dim,1)])]); % changed in dev9
            limits(dim,2) = max([1, min([limits(dim,2), size(target,dim)])]); % changed in dev9
        end
        result = target(limits(1,1):limits(1,2),limits(2,1):limits(2,2),limits(3,1):limits(3,2));
    end

    function [result] = phaseIFFTN(dataSource, deltas)
        % Deltas in pts/K
        for dim = 1:3
            deltas(:,dim) = 0.001*deltas(:,dim)/dataSource.headerInfo.dim{dim}.ppmPerPoint;
        end
        % Dimensions (keeping them here is actually faster than asking for them every time):
        n = size(dataSource.ifftn);
        % Dimensions of the generated part: n/2
        % Function indices same as signs of octants in Cartesian coordinate system:
        % P = (+), N = (-), e.g. pNPN is p(-,+,-) what corresponds to the 6th octant
        pPPP = (zeros(n)); % GPU version: pPPP = gpuArray(zeros(n));
        [pNPP, pNNP, pPNP] = deal(pPPP); % because we obtain pPPN, pNPN, pNNN, pPNN as conjugates later on, no need to assign them now
        % Calculating corrections (doing it here reduces multiplication count in the loop):
        cst_n = 2*pi*1i./n;
        cst = zeros(3,1); % just a preallocation
        for i = 1:(n(1)/2)
            pPPP1 = (zeros(n(2:3))); % GPU version: pPPP1 = gpuArray(zeros(n(2:3)));
            [pNPP1, pNNP1, pPNP1] = deal(pPPP1);
            for kd = 1:numel(dataSource.tempList)
                % Phase can be given as any function of temperature.
                phi = zeros(3, numel(dataSource.tempList(kd)));
                for dim = 1:3
                    phi(dim, :) = cst_n(dim)*(dataSource.tempList(kd) * deltas(1,dim) + dataSource.tempList(kd)^2 * deltas(2,dim));
                end
                cst(1) = exp((i-1)*phi(1));
                for jd = 1:size(dataSource.NUSList,2)
                    j = dataSource.NUSList(2,jd,kd)+1;
                    k = dataSource.NUSList(1,jd,kd)+1;
                    cst(2) = exp((j-1)*phi(2));
                    cst(3) = exp((k-1)*phi(3));
                    pPPP1(j,k) = conj(cst(1)*cst(2)*cst(3)); % pNNN1(j,k) = cst(1)*cst(2)*cst(3); % equal to conj(pPPP1)
                    pNPP1(j,k) = cst(1)*conj(cst(2)*cst(3)); % pPNN1(j,k) = conj(cst(1))*cst(2)*cst(3); % equal to conj(pNPP1)
                    pNNP1(j,k) = cst(1)*cst(2)*conj(cst(3)); % pPPN1(j,k) = conj(cst(1)*cst(2))*cst(3); % equal to conj(pNNP1)
                    pPNP1(j,k) = conj(cst(1))*cst(2)*conj(cst(3)); % pNPN1(j,k) = cst(1)*conj(cst(2))*cst(3); % equal to conj(pPNP1)
                end
            end
            pPPP(i,:,:) = pPPP1; % GPU version: gather(pPPP1);
            pNPP(i,:,:) = pNPP1; % GPU version: gather(pNPP1);
            pNNP(i,:,:) = pNNP1; % GPU version: gather(pNNP1);
            pPNP(i,:,:) = pPNP1; % GPU version: gather(pPNP1);
            % pPPN(i,:,:) = pPPN1; % Conjugating later on is way faster
            % pNPN(i,:,:) = pNPN1; % Conjugating later on is way faster
            % pNNN(i,:,:) = pNNN1; % Conjugating later on is way faster
            % pPNN(i,:,:) = pPNN1; % Conjugating later on is way faster
        end
        pPPN = conj(pNNP); % THIS IS MANDATORY, DO NOT COMMENT OUT!
        pNPN = conj(pPNP); % THIS IS MANDATORY, DO NOT COMMENT OUT!
        pNNN = conj(pPPP); % THIS IS MANDATORY, DO NOT COMMENT OUT!
        pPNN = conj(pNPP); % THIS IS MANDATORY, DO NOT COMMENT OUT!
        % You cannot comment out the four lines above because pPPN, pNPN, pNNN and pPNN are not assigned above!
        pVE = pPPP +... % scheme: flip along dims marked as N, then circshift by 1 point along them (works for all N-dimensional echoes!)
            + circshift(flip(pNPP,1),[1 0 0]) +...
            + circshift(flip(flip(pNNP,1),2),[1 1 0]) +...
            + circshift(flip(pPNP,2),[0 1 0]) +...
            + circshift(flip(pPPN,3),[0 0 1]) +...
            + circshift(flip(flip(pNPN,1),3),[1 0 1]) +...
            + circshift(flip(flip(flip(pNNN,1),2),3),[1 1 1]) +...
            + circshift(flip(flip(pPNN,2),3),[0 1 1]);
        % GPU version:
        % pVE = gpuArray(pPPP +...
        % + circshift(flip(pNPP,1),[1 0 0]) +...
        % + circshift(flip(flip(pNNP,1),2),[1 1 0]) +...
        % + circshift(flip(pPNP,2),[0 1 0]) +...
        % + circshift(flip(pPPN,3),[0 0 1]) +...
        % + circshift(flip(flip(pNPN,1),3),[1 0 1]) +...
        % + circshift(flip(flip(flip(pNNN,1),2),3),[1 1 1]) +...
        % + circshift(flip(flip(pPNN,2),3),[0 1 1]));
        pVE(1,:,:) = pVE(1,:,:)./2;
        pVE(:,1,:) = pVE(:,1,:)./2;
        pVE(:,:,1) = pVE(:,:,1)./2;
        result = real(fftn(dataSource.ifftn .* pVE)); % GPU version: result = gather(real(fftn(dataSource.ifftn .* pVE)));
    end

    function [result] = BruteForceButtonPushed(app, loaded)
        if (size(app.UITable.Data,1) < 1) || (numel(loaded.LastState.SearchedDeltas) < 1)
            error('No data to process; quitting Terminal Radonizer...');
        end
        delOrder = loaded.LastState.SearchedDeltas;
        tableCopy = app.UITable.Data;
        limitTable = tableCopy(:,8:10);
        for dim = 1:3
            limitTable{:,dim} = ppm2pts(app,limitTable{:,dim},dim);
        end
        dataCopy = app.Data;
        % hBox = msgbox('Performing parallel Brute Force for peaks in the table, please wait.','Brute Force in progress','help','modal');
        rowTempFunction = @(arg1,arg2,arg3,arg4)rowBF(arg1,arg2,arg3,arg4);
        parfor row = 1:size(app.UITable.Data,1)
            tableCopy(row,:) = feval(rowTempFunction, tableCopy(row,:), dataCopy, limitTable(row,:), delOrder);
        end
        result = tableCopy;
        % app.UITable.Data = tableCopy;
        % close(hBox);
        
        function [results] = rowBF(tableRow, dataSource, limitRow, delOrder)
            deltas = [tableRow.dL1(1), tableRow.dL2(1), tableRow.dL3(1); tableRow.dQ1(1), tableRow.dQ2(1), tableRow.dQ3(1)];
            limits = [limitRow.sX(1,:); limitRow.sY(1,:); limitRow.sZ(1,:)];
            for dim = delOrder
                [d,p] = ind2sub([3,2], dim);
                gridParams = tableRow{1,18+2*(d-1)+p};
                grid = gridParams(1):gridParams(3):gridParams(2);
                maxes = nan(1,numel(grid));
                tempDeltas = deltas;
                for id = 1:numel(grid)
                    % TO DO: calculate only if grid point is within range
                    tempDeltas(d,p) = grid(id);
                    maxes(id) = max(trim3D(phaseIFFTN(dataSource, tempDeltas),limits),[],'all');
                end
                [~, maxID] = max(maxes,[],'all','linear');
                deltas(d,p) = grid(maxID);
            end
            tableRow.dL1(1) = deltas(1,1);
            tableRow.dL2(1) = deltas(2,1);
            tableRow.dL3(1) = deltas(3,1);
            tableRow.dQ1(1) = deltas(1,2);
            tableRow.dQ2(1) = deltas(2,2);
            tableRow.dQ3(1) = deltas(3,2);
            tableRow.maxF(1) = maxes(maxID);
            results = tableRow;
        end
    end

    function [result] = ParticleSwarmButtonPushed(app,loaded)
        if size(app.UITable.Data,1) < 1
            error('Data table empty; quitting Terminal Radonizer...');
        end
        tableCopy = app.UITable.Data;
        limitTable = tableCopy(:,8:10);
        for dim = 1:3
            limitTable{:,dim} = ppm2pts(app,limitTable{:,dim},dim);
        end
%         dataCopy = app.Data; % better avoid that due to memory issues, it holds some huge variablesw
%         hBar = waitbar(0,'Performing parallel Particle Swarm for peaks in the table, please wait.','Name','Particle Swarm in progress','WindowStyle','modal');
        if loaded.LastState.LimitedSwarm == 1
%             distances = [app.dQ1SwarmRangeEditField.Value, app.dQ2SwarmRangeEditField.Value, app.dQ3SwarmRangeEditField.Value];
            for row = 1:size(app.UITable.Data,1)
                tableCopy(row,:) = rowSwarmLimQ(loaded, tableCopy(row,:), app.Data, limitTable(row,:), loaded.LastState.LimitedSwarmParams);
%                 if isgraphics(hBar) % may have been closed by user, added in dev11
%                     waitbar(row/size(app.UITable.Data,1),hBar);
%                 end
            end
        else
            for row = 1:size(app.UITable.Data,1)
                tableCopy(row,:) = rowSwarm(loaded, tableCopy(row,:), app.Data, limitTable(row,:));
%                 if isgraphics(hBar) % may have been closed by user, added in dev11
%                     waitbar(row/size(app.UITable.Data,1),hBar);
%                 end
            end
        end
        result = tableCopy;
        % app.UITable.Data = tableCopy;
        % close(hBar);

        function [results] = rowSwarm(loaded, tableRow, dataSource, limitRow)
            limits = [limitRow.sX(1,:); limitRow.sY(1,:); limitRow.sZ(1,:)];
            kS = 3; % size(limits,1);

            lb = [[tableRow.gL1(1,1), tableRow.gL2(1,1), tableRow.gL3(1,1)]; [tableRow.gQ1(1,1), tableRow.gQ2(1,1), tableRow.gQ3(1,1)]]; % lower bounds
            ub = [[tableRow.gL1(1,2), tableRow.gL2(1,2), tableRow.gL3(1,2)]; [tableRow.gQ1(1,2), tableRow.gQ2(1,2), tableRow.gQ3(1,2)]]; % upper bounds

            % tic;
            func = @(d)1/max(abs(trim3D(phaseIFFTN(dataSource, reshape(d, [2, kS])), limits)),[],'all'); % inverse for minimization

            options.m = kS*2; % number of variables
            options.n = loaded.LastState.SwarmParams(1); % for population size - in each dimension
            options.maxiter = loaded.LastState.SwarmParams(2); % set maximum number of iteration
            options.maxrun = loaded.LastState.SwarmParams(3); % set maximum number of runs need to be
            options.wmax = 0.9; % max inertia weight
            options.wmin = 0.9; % min inertia weight
            options.c1 = 1; % acceleration factor 1
            options.c2 = 2; % acceleration factor 2
            options.tol = 1/4e20; % default: 1/1.49e4
            % options.d = 0.05; % distance from edge - for initialization
            [x_opt, f_opt] = pso(func, lb, ub, options);
            % toc;
            deltas = reshape(x_opt,[2, kS])';
            % % now this obviously would have to be reworked if number of deltas changed, it has nothing to do with calculations themselves
            tableRow.dL1(1) = deltas(1,1);
            tableRow.dL2(1) = deltas(2,1);
            tableRow.dL3(1) = deltas(3,1);
            tableRow.dQ1(1) = deltas(1,2);
            tableRow.dQ2(1) = deltas(2,2);
            tableRow.dQ3(1) = deltas(3,2);
            tableRow.maxF(1) = 1/f_opt;
            results = tableRow;

            function [rfgbest, fffmin] = pso(fun, LB, UB, options)
                m = options.m; % number of variables
                n = options.n; % population size
                wmax = options.wmax; % inertia weight
                wmin = options.wmin; % inertia weight
                c1 = options.c1; % acceleration factor
                c2 = options.c2; % acceleration factor
                maxite = options.maxiter; % set maximum number of iteration
                maxrun = options.maxrun; % set maximum number of runs
                tol = options.tol;
                % % d = 0.05; % options.d;

                % % pso initialization----------------------------------------------start
                % disp('Performing initial steps... ');
                % tic
                % % C = {};
                x0 = nan(n,m); % preallocating for speed
                for i = 1:n
                    for j = 1:m
                        % % x0(i,j) = round(LB(j) + rand()*(UB(j) - LB(j)));
                        x0(i,j) = LB(j) + rand()*(UB(j) - LB(j));
                        % % C{end+1} = linspace(LB(j) + d*(UB(j) - LB(j)), UB(j) - d*(UB(j) - LB(j)), n);
                    end
                end
                % % D = C;
                % % [D{:}] = ndgrid(C{:});
                % % x0 = cell2mat(cellfun(@(m)m(:), D, 'uni', 0));
                % % disp(sprintf('Population size: %i', num2str(size(x0, 1))))
                % % clear('C','D');
                % toc
                % disp('...done!');
                fff = nan(1,maxrun); % preallocating for speed
                rgbest = nan(maxrun,m); % preallocating for speed
                for run = 1:maxrun
                    % disp(sprintf('Starting algorithm run: %i (out of %i max)', run, maxrun));
                    x = x0; % initial population, n x m
                    v = 0.1 * x0; % initial velocity
                    % % if parpool('local') == 0 % checking to see if my pool is already open
                    % %     parpool open 2
                    % % end
                    % % % parDQ = parallel.pool.DataQueue;
                    % % % listener = afterEach(parDQ, @(i) disp(sprinf('i: %i', i)));
                    % disp('Initialization... ');
                    % tic
                    parfor i = 1:n
                        f0(i,1) = fun(x0(i,:));
                    end
                    % disp('...done!');
                    % toc
                    % % % delete(listener);
                    [fmin0, index0] = min(f0);
                    pbest = x0; % initial pbest
                    gbest = x0(index0,:); % initial gbest
                    % % pso initialization------------------------------------------------end

                    % % f = fopen('test.txt','a');
                    % % fprintf(f, '%4f %4f %4f %4f %6e\n', [pbest, f0]);
                    % % fclose(f);

                    % % pso algorithm---------------------------------------------------start
                    % disp('!-- PSO algorithm started. ');
                    ite = 1;
                    % ffmin = zeros(maxite,1); % preallocating for speed
                    % % tolerance = 1;
                    % % while ite <= maxite && tolerance > tol
                    while ite <= maxite && fmin0 > tol
                        % disp(sprintf('Entering WHILE loop, iteration %i (out of %i max)', ite, maxite));
                        w = wmax - (wmax - wmin) * ite/maxite; % update inertial weight
                        % disp('Updating PSO velocity... ');
                        % tic
                        parfor i=1:n
                            for j=1:m
                                v(i,j) = w*v(i,j) + c1*rand()*(pbest(i,j) - x(i,j)) + c2*rand()*(gbest(1,j) - x(i,j));
                                % % v(i,j) = w*v(i,j) + c1*(pbest(i,j) - x(i,j)) + c2*(gbest(1,j) - x(i,j));
                            end
                        end
                        % disp('...done!');
                        % toc
                        % disp('PSO position update... ');
                        % tic
                        parfor i = 1:n
                            for j = 1:m
                                x(i,j) = x(i,j)+v(i,j);
                            end
                        end
                        % disp('...done!');
                        % toc
                        % disp('Handling boundary violations... ');
                        % tic
                        % vv = 0;
                        parfor i = 1:n % not sure if parfor is any faster due to broadcast variable LB
                            for j = 1:m
                                if x(i,j) < LB(j)
                                    x(i,j) = LB(j);
                                    % vv = vv + 1;
                                elseif x(i,j) > UB(j)
                                    x(i,j) = UB(j);
                                    % vv = vv + 1;
                                end
                            end
                        end
                        % disp(sprintf('vv: %i', vv));
                        % disp('...done!');
                        % toc
                        % disp('Evaluating fitness... ')
                        % % % listener = afterEach(parDQ, @(i) disp(sprintf('beginning loop no. %i',i)));
                        % tic
                        parfor i = 1:n
                            f(i,1) = fun(x(i,:));
                        end
                        % disp('...done!');
                        % % % delete(listener);
                        % toc
                        % disp('Updating pbest and fitness... ');
                        % tic
                        % k = 0;
                        parfor i = 1:n
                            if f(i,1) < f0(i,1)
                                % k = k + 1;
                                pbest(i,:) = x(i,:);
                                f0(i,1) = f(i,1);
                            end
                        end
                        % disp(sprintf('k: %i', k));
                        % disp('...done!');
                        % toc
                        [fmin, index] = min(f0); % finding out the best particle
                        % ffmin(ite) = fmin; % storing best fitness
                        % % ffite(run) = ite; % storing iteration count
                        % % updating gbest and best fitness
                        if fmin < fmin0
                            gbest = pbest(index,:);
                            fmin0 = fmin;
                        end
                        % % % calculating tolerance
                        % % if ite > 100;
                        % %    tolerance = abs(ffmin(ite-100,run)-fmin0);
                        % % end
                        % disp(sprintf('Iteration %i results: ', ite));
                        % disp(sprintf('Best particle (index): %i, Objective fun (fmin0): %8.4f', index, fmin0));
                        % disp(['Deltas: ', num2str(gbest,'%4f ')]);
                        % disp(sprintf('MaxAfter: %8g', 1/fmin0));
                        % disp(sprintf('fmin: %.8g', fmin));
                        ite = ite + 1;
                        % % f = fopen('gbest.txt','a');
                        % % fprintf(f, '%4f %4f %4f %4f %6e\n', [gbest, fmin0]);
                        % % fclose(f);
                    end
                    % ffmin = ffmin(1:ite); % trimming excess elements
                    % disp('!-- PSO algorithm ended.');
                    % % pso algorithm-----------------------------------------------------end
                    fvalue = fun(gbest);
                    fff(run) = fvalue;
                    rgbest(run,:) = gbest;
                    if fmin0 <= tol
                        % disp('Search threshold reached!');
                        break
                    end
                    % disp(newline);
                end
                [fffmin, indexff] = min(fff); % finding out the best run
                % disp(sprintf('Best run: %i, fffmin: %.8g', indexff, fffmin));
                rfgbest = rgbest(indexff,:);
            end
        end

        function [results] = rowSwarmLimQ(loaded, tableRow, dataSource, limitRow, distances)
            limits = [limitRow.sX(1,:); limitRow.sY(1,:); limitRow.sZ(1,:)];

            Tm = max(loaded.AppData.tempList,[],'all'); % maximum temperature
            k = [tableRow.dL1, tableRow.dL2, tableRow.dL3]; % previous linear results
            kS = 3; % equal to numel(k);
            D = logical(loaded.LastState.LimitedSwarmDeltas);
            if loaded.LastState.LimitedSwarmType == '1'
                D = abs(Tm*(k .* D));
            else
                D = abs(distances .* D);
            end
            % tic;
            func = @(d)1/max(abs(trim3D(phaseIFFTN(dataSource, reshape(d, [2, kS])), limits)),[],'all'); % inverse for minimization

            % options.m = numel(k)*2; % number of variables
            options.n = loaded.LastState.SwarmParams(1); % for population size - in each dimension
            options.maxiter = loaded.LastState.SwarmParams(2); % set maximum number of iteration
            options.maxrun = loaded.LastState.SwarmParams(3); % set maximum number of runs need to be
            options.wmax = 0.9; % max inertia weight
            options.wmin = 0.9; % min inertia weight
            options.c1 = 1; % acceleration factor 1
            options.c2 = 2; % acceleration factor 2
            options.tol = 1/(4e20); % 1/height; default: 1/(4e8)
            % options.d = 0.05; % distance from edge - for initialization
            [x_opt, f_opt] = pso_limQ(func, D, Tm, k, options);
            % toc;
            deltas = reshape(x_opt,[2, kS])';
            % % now this obviously would have to be reworked if number of deltas changed, it has nothing to do with calculations themselves
            tableRow.dL1(1) = deltas(1,1);
            tableRow.dL2(1) = deltas(2,1);
            tableRow.dL3(1) = deltas(3,1);
            tableRow.dQ1(1) = deltas(1,2);
            tableRow.dQ2(1) = deltas(2,2);
            tableRow.dQ3(1) = deltas(3,2);
            tableRow.maxF(1) = 1/f_opt;
            results = tableRow;

            function [rfgbest, fffmin] = pso_limQ(fun, D, Tm, k, options)
                m = numel(k)*2;
                n = options.n; % population size
                wmax = options.wmax; % inertia weight
                wmin = options.wmin; % inertia weight
                c1 = options.c1; % acceleration factor
                c2 = options.c2; % acceleration factor
                maxite = options.maxiter; % set maximum number of iteration
                maxrun = options.maxrun; % set maximum number of runs need to be
                tol = options.tol; % default: 1/(4e8)

                bprimA = 2*D*(1+sqrt(2))/Tm; % D can be a vector! k as well
                % bprimB = 2*D/Tm; % apparently it remains unused

                LB_lin = -bprimA + k;
                UB_lin = bprimA + k;

                fff = nan(1,maxrun); % preallocating for speed
                rgbest = nan(maxrun,m); % preallocating for speed
                for run = 1:maxrun
                    % disp(sprintf('Starting algorithm run: %i (out of %i max)', run, maxrun));
                    % disp('Calculating bounds... ');
                    % tic
                    % % pso initialization----------------------------------------------start
                    x0 = nan(n,m); % preallocating for speed
                    for i = 1:n % particles
                        for j = 1:(m/2) % j = (1:2:m) % dimensions, linear
                            % % rng('shuffle'); % ???
                            % %  Ceils are expensive, avoid them in loops! Changed in dev11
                            % % x0(i,j) = LB_lin(ceil(j/2)) + rand()*(UB_lin(ceil(j/2)) - LB_lin(ceil(j/2)));
                            x0(i,2*j-1) = LB_lin(j) + rand()*(UB_lin(j) - LB_lin(j));
                            % % LB_sq = calc_LB_sq(x0(i,j)-k(ceil(j/2)), Tm, D(ceil(j/2)));
                            LB_sq = calc_LB_sq(x0(i,2*j-1)-k(j), Tm, D(j));
                            % % UB_sq = calc_UB_sq(x0(i,j)-k(ceil(j/2)), Tm, D(ceil(j/2)));
                            UB_sq = calc_UB_sq(x0(i,2*j-1)-k(j), Tm, D(j));
                            % % x0(i,j+1) = LB_sq + rand()*(UB_sq - LB_sq);
                            x0(i,2*j) = LB_sq + rand()*(UB_sq - LB_sq);
                        end
                    end
                    % disp('...done!');
                    % toc
                    x = x0; % initial population, n x m
                    % % if parpool('local') == 0 % checking to see if my pool is already open
                    % %     parpool open 2
                    % % end
                    % disp('Initialization... ');
                    % tic
                    parfor i = 1:n
                        f0(i,1) = fun(x0(i,:));
                    end
                    [fmin0, index0] = min(abs(f0));
                    pbest = x0; % initial pbest
                    gbest = x0(index0,:); % initial gbest
                    % disp(sprintf('Best deltas: %f %f %f %f %f %f', gbest(:)));
                    % disp(sprintf('MaxAfter: %8e', 1/fmin0));
                    % disp('...done!');
                    % toc

                    % % pso algorithm---------------------------------------------------start
                    % disp(sprintf('!-- PSO algorithm started. '));
                    v = 0.1*x0;
                    ite = 1;
                    % ffmin = zeros(maxite,1); % preallocating for speed
                    while ite <= maxite && abs(fmin0) > tol
                        % disp(sprintf('Entering WHILE loop, iteration %i (out of %i max)', ite, maxite));
                        w = wmax-(wmax-wmin)*ite/maxite; % update inertial weight
                        % disp('Updating PSO velocity... ');
                        % tic
                        parfor i = 1:n
                            for j = 1:m
                                % % rng('shuffle'); % ???!!!
                                v(i,j) = w*v(i,j) + c1*rand()*(pbest(i,j)-x(i,j)) + c2*rand()*(gbest(1,j)-x(i,j));
                            end
                        end
                        % disp('...done!');
                        % toc
                        % disp('PSO position update... ');
                        % tic
                        parfor i = 1:n
                            for j = 1:m
                                x(i,j) = x(i,j) + v(i,j);
                            end
                        end
                        % disp('...done!');
                        % toc
                        % disp('Handling boundary violations... ');
                        % tic
                        % vv = 0;
                        % % parfor i = 1:n % cannot be simultaneous in that form!
                        for i = 1:n
                            for j = 1:(m/2) % j = (1:2:m) % dimensions, linear
                                % %  Ceils are expensive, avoid them in loops! Changed in dev11
                                if x(i,2*j-1) < LB_lin(j) % if x(i,j) < LB_lin(ceil(j/2))
                                    % % x(i,j) = LB_lin(ceil(j/2));
                                    x(i,2*j-1) = LB_lin(j);
                                    % vv = vv + 1;
                                elseif x(i,2*j-1) > UB_lin(j) % x(i,j) > UB_lin(ceil(j/2))
                                    % % x(i,j) = UB_lin(ceil(j/2));
                                    x(i,2*j-1) = UB_lin(j);
                                    % vv = vv + 1;
                                end
                                % % LB_sq = calc_LB_sq(x(i,j)-k(ceil(j/2)), Tm, D(ceil(j/2)));
                                LB_sq = calc_LB_sq(x(i,2*j-1)-k(j), Tm, D(j));
                                % % UB_sq = calc_UB_sq(x(i,j)-k(ceil(j/2)), Tm, D(ceil(j/2)));
                                UB_sq = calc_UB_sq(x(i,2*j-1)-k(j), Tm, D(j));
                                if x(i,2*j) < LB_sq % x(i,j+1) < LB_sq
                                    % % x(i,j+1) = LB_sq;
                                    x(i,2*j) = LB_sq;
                                    % vv = vv + 1;
                                elseif x(i,2*j) > UB_sq % x(i,j+1) > UB_sq
                                    % % x(i,j+1) = UB_sq;
                                    x(i,2*j) = UB_sq;
                                    % vv = vv + 1;
                                end
                            end
                        end
                        % disp('...done!');
                        % toc
                        % disp('Evaluating fitness... ');
                        % tic
                        parfor i = 1:n
                            f(i,1) = fun(x(i,:));
                        end
                        % disp('...done!');
                        % toc
                        % disp('Updating pbest and fitness... ');
                        % tic
                        % % k1 = 0;
                        parfor i = 1:n
                            if f(i,1) < f0(i,1)
                                % % k1 = k1 + 1;
                                pbest(i,:) = x(i,:);
                                f0(i,1) = f(i,1);
                            end
                        end
                        % disp('...done!');
                        % toc
                        [fmin,index] = min(abs(f0)); % finding out the best particle
                        % ffmin(ite) = fmin; % storing best fitness
                        % % ffite(run) = ite; % storing iteration count
                        % % updating gbest and best fitness
                        if fmin < fmin0
                            gbest = pbest(index,:);
                            fmin0 = fmin;
                        end
                        ite = ite + 1;
                        % disp(sprintf('Iteration %i results: ', ite));
                        % disp(sprintf('Best particle (index): %i, Objective fun (fmin0): %8.4f', index, fmin0));
                        % disp(['Deltas: ', num2str(gbest,'%4f ')]);
                        % disp(sprintf('MaxAfter: %8g', 1/fmin0));
                        % disp(sprintf('gbest: %.8g, fmin: %.8g', gbest, fmin));
                        % % f = fopen('gbest.txt','a');
                        % % fprintf(f, '%4f %4f %4f %4f %4f %4f %6e\n', [gbest, fmin0]);
                        % % fclose(f);
                    end
                    % ffmin = ffmin(1:ite); % trimming excess elements
                    % disp('!-- PSO algorithm ended.');
                    % % pso algorithm-----------------------------------------------------end
                    fvalue = fun(gbest);
                    fff(run) = fvalue;
                    rgbest(run,:) = gbest;
                    if abs(fmin0) <= tol
                        % disp('Desired peak height reached!');
                        break
                    end
                    % disp(newline);
                end
                [fffmin,indexff] = min(abs(fff)); % finding out the best run
                % disp(sprintf('Best run: %i, fffmin: %.8g', indexff, fffmin));
                rfgbest = rgbest(indexff,:);

                function [LB_sq] = calc_LB_sq(bprim, Tm, D)
                    if D ~= 0
                        if bprim > 0
                            LB_sq = -bprim/Tm - D/(Tm^2);
                        elseif bprim <= 0 && bprim > - 2*D/Tm % |bprimB|
                            LB_sq = -bprim/(2*Tm);
                        else % brim <= -bprimB
                            LB_sq = bprim^2/(4*D);
                        end
                    else
                        LB_sq = 0;
                    end
                end

                function [UB_sq] = calc_UB_sq(bprim, Tm, D)
                    if D ~= 0
                        if bprim <= 0
                            UB_sq = -bprim/Tm + D/(Tm^2);
                        elseif bprim > 0 && bprim <= 2*D/Tm % |bprimB|
                            UB_sq = -bprim/(2*Tm);
                        else % from bprimB to bprimA
                            UB_sq = -bprim^2/(4*D);
                        end
                    else
                        UB_sq = 0;
                    end
                end
            end
        end
    end
end